/****************************************************************************
* VCGLib                                                            o o     *
* Visual and Computer Graphics Library                            o     o   *
*                                                                _   O  _   *
* Copyright(C) 2004-2012                                           \/)\/    *
* Visual Computing Lab                                            /\/|      *
* ISTI - Italian National Research Council                           |      *
*                                                                    \      *
* All rights reserved.                                                      *
*                                                                           *
* This program is free software; you can redistribute it and/or modify      *
* it under the terms of the GNU General Public License as published by      *
* the Free Software Foundation; either version 2 of the License, or         *
* (at your option) any later version.                                       *
*                                                                           *
* This program is distributed in the hope that it will be useful,           *
* but WITHOUT ANY WARRANTY; without even the implied warranty of            *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
* GNU General Public License (http://www.gnu.org/licenses/gpl.txt)          *
* for more details.                                                         *
*                                                                           *
****************************************************************************/

#include <vcg/complex/complex.h>
#include <wrap/io_trimesh/import_ply.h>
#include <wrap/io_trimesh/export_ply.h>
#include <vcg/complex/algorithms/update/color.h>

using namespace vcg;
using namespace std;


class MyEdge;
class MyFace;
class MyVertex;
struct MyUsedTypes : public vcg::UsedTypes<	vcg::Use<MyVertex>   ::AsVertexType,
                                        vcg::Use<MyEdge>     ::AsEdgeType,
                                        vcg::Use<MyFace>     ::AsFaceType>{};

class MyVertex  : public vcg::Vertex<MyUsedTypes, vcg::vertex::Coord3f, vcg::vertex::Normal3f, vcg::vertex::BitFlags  >{};
class MyFace    : public vcg::Face< MyUsedTypes,  vcg::face::Normal3f ,vcg::face::FFAdj,  vcg::face::Color4b, vcg::face::VertexRef, vcg::face::Qualityf, vcg::face::BitFlags > {};
class MyEdge    : public vcg::Edge<MyUsedTypes>{};
class MyMesh    : public vcg::tri::TriMesh< std::vector<MyVertex>, std::vector<MyFace> , std::vector<MyEdge>  > {};

/*
 * Smooth the boundary of the labyrinth
 *
 * Iterate on the faces for each "boundary" edge
 * and accumulates each vertex of the edge on the other vertex.
 *
 */

void SmoothLabyrinth(MyMesh &m)
{
  MyMesh::PerVertexAttributeHandle<Point3f> accH =
      tri::Allocator<MyMesh>::GetPerVertexAttribute<Point3f>(m,"vertAccum");

  for(MyMesh::VertexIterator vi=m.vert.begin();vi!=m.vert.end();++vi)
    accH[vi]=vi->P()*4.0f;

  for(MyMesh::FaceIterator fi=m.face.begin();fi!=m.face.end();++fi)
  {
    for(int i=0;i<3;++i)
    {
     if(fi->IsV() && !fi->FFp(i)->IsV())
     {
      accH[fi->V0(i)]+= fi->V1(i)->P();
      accH[fi->V1(i)]+= fi->V0(i)->P();
     }
    }
  }

  for(MyMesh::VertexIterator vi=m.vert.begin();vi!=m.vert.end();++vi)
    if(vi->IsV())
      vi->P() = accH[vi] / 6.0f;
}

// Questa funzione deve generare le pareti verticali del labirinto.
// piu precisamente per ogni edge del bottom deve formare il quadrilatero formato
// da due triangoli che connette tale edge con il corrispondente edge della parte top del labirinto.
void GenerateSideMesh(MyMesh &m, MyMesh &side, float thr)
{
  for(MyMesh::FaceIterator fi=m.face.begin();fi!=m.face.end();++fi)
  {
    for(int i=0;i<3;++i)
    {
     if(fi->IsV() && !fi->FFp(i)->IsV())
     {
       Point3f P0B ; // Da inizializzare alle giuste posizioni....
       Point3f P1B;
       Point3f P0T;
       Point3f P1T;

       tri::Allocator<MyMesh>::AddFace(P0B,P1B,P0T); // Da controllare!!!
       tri::Allocator<MyMesh>::AddFace(P0B,P1T,P0T);
     }
    }
  }
}

int main(int argc, char **argv)
{
  MyMesh m;
  tri::io::ImporterPLY<MyMesh>::Open(m,argv[1]);

  printf("Read mesh %i %i\n",m.FN(),m.VN());
  tri::UpdateTopology<MyMesh>::FaceFace(m);
  tri::UpdateNormal<MyMesh>::PerVertexNormalizedPerFaceNormalized(m);
  tri::UpdateFlags<MyMesh>::FaceClearV(m);
  tri::UpdateFlags<MyMesh>::VertexClearV(m);
  tri::UpdateColor<MyMesh>::PerFaceConstant(m,Color4b::Black);
  int faceCnt=0;
  vector<MyFace*> faceStack;
  faceStack.push_back(&m.face[0]);
  m.face[0].SetV();
  for(int i=0;i<3;++i) m.face[0].V(i)->SetV();
  while(!faceStack.empty())
  {
    if(faceStack.size()>1)
      swap(faceStack.back(),faceStack[rand()%(faceStack.size()-1)]);
    MyFace *fp = faceStack.back();
    faceStack.pop_back();
    for(int i=0;i<3;++i)
    {MyFace *curfp = fp->FFp(i);
      if(!curfp->IsV())
      {
        bool badFace=false;
        int visitedVertexCnt=0;
        for(int j=0;j<3;++j)
        {
          if(curfp->FFp(j)->IsV() && curfp->FFp(j) != fp)
            badFace=true;
          if(curfp->V(j)->IsV()) ++visitedVertexCnt;
        }

        if(!badFace && visitedVertexCnt==2)
        {
          curfp->SetV();
          curfp->SetS();
          for(int j=0;j<3;++j)
            curfp->V(j)->SetV();
          ++faceCnt;
          curfp->C() = Color4b::ColorRamp(0,m.fn/10,(faceCnt)%(m.fn/10));
          curfp->Q() = faceCnt;
          faceStack.push_back(curfp);
        }
      }
    }
  }
  printf("Visited %i faces \n",faceCnt);
  tri::io::ExporterPLY<MyMesh>::Save(m,"out.ply",tri::io::Mask::IOM_FACECOLOR | tri::io::Mask::IOM_FACEQUALITY);
  SmoothLabyrinth(m);
  SmoothLabyrinth(m);
  SmoothLabyrinth(m);
  tri::io::ExporterPLY<MyMesh>::Save(m,"out2.ply",tri::io::Mask::IOM_FACECOLOR | tri::io::Mask::IOM_FACEQUALITY);
  tri::io::ExporterPLY<MyMesh>::Save(m,"out2.ply",tri::io::Mask::IOM_FACECOLOR | tri::io::Mask::IOM_FACEQUALITY);
  tri::io::ExporterPLY<MyMesh>::Save(m,"out2.ply",tri::io::Mask::IOM_FACECOLOR | tri::io::Mask::IOM_FACEQUALITY);
  MyMesh bottom,top;
  tri::Append<MyMesh,MyMesh>::MeshCopy(bottom,m,true);
  tri::UpdateSelection<MyMesh>::FaceInvert(m);
  tri::Append<MyMesh,MyMesh>::MeshCopy(top,m,true);


  float thr = m.bbox.Diag()/100.0f;
  GenerateSideMesh(m, side, thr);
  for(MyMesh::VertexIterator vi=top.vert.begin();vi!=top.vert.end();++vi)
    vi->P() = vi->P() + vi->N()*thr;

  tri::io::ExporterPLY<MyMesh>::Save(top,"top.ply",tri::io::Mask::IOM_FACECOLOR | tri::io::Mask::IOM_FACEQUALITY);
  tri::io::ExporterPLY<MyMesh>::Save(bottom,"bottom.ply",tri::io::Mask::IOM_FACECOLOR | tri::io::Mask::IOM_FACEQUALITY);

  return 0;
}
